/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "eb.h"
#include "error.h"
#include "io.h"
#include "language.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX	MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX	1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */


/*
 * Return the number of languages int the book.
 */
int
eb_language_count(book)
    EB_Book *book;
{
    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return -1;
    }

    return book->langcount;
}


/*
 * Make a list of languages the book has.
 */
int
eb_language_list(book, list)
    EB_Book *book;
    EB_Language_Code *list;
{
    EB_Language *lng;
    int i;

    /*
     * Language information in the book must have been initalized.
     */
    if (book->languages == NULL) {
	eb_error = EB_ERR_NO_LANG;
	return -1;
    }

    for (i = 0, lng = book->languages;
	 i < book->langcount; i++, list++, lng++)
	*list = lng->code;

    return book->langcount;
}


/*
 * Test that the book supports a language `code'.
 */
int
eb_have_language(book, code)
    EB_Book *book;
    EB_Language_Code code;
{
    EB_Language *lng;
    int i;

    /*
     * Language information in the book must have been loaded.
     */
    if (book->languages == NULL) {
	eb_error = EB_ERR_NO_LANG;
	return 0;
    }

    /*
     * Find the language entry.
     */
    for (i = 0, lng = book->languages; i < book->langcount; i++, lng++) {
	if (lng->code == code)
	    return 1;
    }

    eb_error = EB_ERR_NO_SUCH_LANG;
    return 0;
}


/*
 * Return the current language code.
 */
EB_Language_Code
eb_language(book)
    EB_Book *book;
{
    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return -1;
    }

    return book->langcurrent->code;
}


/*
 * Return a name of the current language.
 */
const char *
eb_language_name(book)
    EB_Book *book;
{
    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return NULL;
    }

    return book->langcurrent->name;
}


/*
 * Return a language name of the number `code'.
 */
const char *
eb_language_name2(book, code)
    EB_Book *book;
    EB_Language_Code code;
{
    EB_Language *lng;
    int i;

    /*
     * Language information in the book must have been initialized.
     */
    if (book->languages == NULL) {
	eb_error = EB_ERR_NO_LANG;
	return NULL;
    }

    /*
     * Find the language entry.
     */
    for (i = 0, lng = book->languages; i < book->langcount; i++, lng++) {
	if (lng->code == code)
	    return lng->name;
    }

    eb_error = EB_ERR_NO_SUCH_LANG;
    return NULL;
}


/*
 * Set a language `code' as the current language.
 */
int
eb_set_language(book, code)
    EB_Book *book;
    EB_Language_Code code;
{
    EB_Language *lng;
    char *msg;
    char language[PATH_MAX + 1];
    int file = -1;
    int i;

    /*
     * Language information in the book must have been initalized,
     * and memories for messages must have been alloated.
     */
    if (book->languages == NULL) {
	eb_error = EB_ERR_NO_LANG;
	goto failed;
    }

    /*
     * If the language to set is the current languages, there is nothing
     * to be done.
     */
    if (book->langcurrent != NULL && book->langcurrent->code == code)
	return book->langcurrent->msgcount;

    /*
     * Search a subbook which has language code `code'.
     */
    for (i = 0, lng = book->languages; i < book->langcount; i++, lng++) {
	if (lng->code == code)
	    break;
    }
    if (book->langcount <= i) {
	eb_error = EB_ERR_NO_SUCH_LANG;
	goto failed;
    }

    /*
     * Allocate memories for message buffer.
     */
    if (book->messages == NULL && eb_initialize_messages(book) < 0)
	goto failed;

    /*
     * Open the language file.
     * Length of the book path and subdir must be checked by caller.
     */
    sprintf(language, "%s/%s", book->path, EB_FNAME_LANGUAGE);
    eb_fix_filename(book, language);
    file = open(language, O_RDONLY);
    if (file < 0) {
	eb_error = EB_ERR_FAIL_OPEN_LANG;
	goto failed;
    }

    /*
     * Read messages.
     * Appends '\0' to each message.
     */
    if (lseek(file, lng->offset, SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_LANG;
	goto failed;
    }
    for (i = 0, msg = book->messages;
	 i < lng->msgcount; i++, msg += EB_MAXLEN_MESSAGE + 2) {
	if (read(file, msg, EB_MAXLEN_MESSAGE + 1) != EB_MAXLEN_MESSAGE + 1) {
	    eb_error = EB_ERR_FAIL_READ_LANG;
	    goto failed;
	}
	*(msg + EB_MAXLEN_MESSAGE + 1) = '\0';
    }

    /*
     * Convert messages from SJIS to JIS when language is Japanese.
     */
    if (lng->code == EB_LANG_JAPANESE) {
        for (i = 0, msg = book->messages;
	     i < lng->msgcount; i++, msg += EB_MAXLEN_MESSAGE + 2)
	    eb_sjis_to_euc(msg + 1, msg + 1);
    }
    book->langcurrent = lng;

    /*
     * Close the language file.
     */
    close(file);

    return lng->msgcount;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file)
	close(file);
    book->langcurrent = NULL;
    return -1;
}


/*
 * Unset the current language.
 */
void
eb_unset_language(book)
    EB_Book *book;
{
    book->langcurrent = NULL;
}


/*
 * Return the number of messaegs in the current language.
 */
int
eb_message_count(book)
    EB_Book *book;
{
    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return -1;
    }

    return book->langcurrent->msgcount;
}


/*
 * Get a list of messages in the current language.
 */
int
eb_message_list(book, list)
    EB_Book *book;
    EB_Message_Code *list;
{
    char *msg;
    int i;

    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return -1;
    }

    /*
     * Make messages list.
     */
    for (i = 0, msg = book->messages; i < book->langcurrent->msgcount;
	 i++, list++, msg += EB_MAXLEN_MESSAGE + 2)
	*list = eb_uint1(msg);

    return book->langcurrent->msgcount;
}


/*
 * Test whether the message `code' is exists in the current language.
 */
int
eb_have_message(book, code)
    EB_Book *book;
    EB_Message_Code code;
{
    char *msg;
    int c;
    int i;
    
    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return 0;
    }
    
    for (i = 0, msg = book->messages; i < book->langcurrent->msgcount;
	 i++, msg += EB_MAXLEN_MESSAGE + 2) {
	c = eb_uint1(msg);
	if (c == code)
	    return 1;
    }

    eb_error = EB_ERR_NO_SUCH_MSG;
    return 0;
}


/*
 * Get the specified message in the current language.
 */
const char *
eb_message(book, code)
    EB_Book *book;
    EB_Message_Code code;
{
    char *msg;
    int c;
    int i;
    
    /*
     * Current language must have been set.
     */
    if (book->langcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_LANG;
	return NULL;
    }
    
    for (i = 0, msg = book->messages; i < book->langcurrent->msgcount;
	 i++, msg += EB_MAXLEN_MESSAGE + 2) {
	c = eb_uint1(msg);
	if (c == code)
	    return msg + 1;
    }

    eb_error = EB_ERR_NO_SUCH_MSG;
    return NULL;
}


