/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#include "eb.h"
#include "error.h"


/*
 * Unexported functions.
 */
#ifdef __STDC__
static EB_Word_Code eb_fix_word_ltn(char *, const char *);
static EB_Word_Code eb_fix_word_jis(char *, const char *);
static void eb_canonicalize_word_jis(char *, const char *);
static void eb_reverse_word_ltn(char *);
static void eb_reverse_word_jis(char *);
#else /* not __STDC__ */
static EB_Word_Code eb_fix_word_ltn();
static EB_Word_Code eb_fix_word_jis();
static void eb_canonicalize_word_jis();
static void eb_reverse_word_ltn();
static void eb_reverse_word_jis();
#endif /* not __STDC__ */


/*
 * Make a fixed word and a cannonicalized word for `WORD SEARCH'.
 *
 * If `inputword' is a KANA word,  EB_WORD_KANA is returned.
 * If `inputword' is a alphabetic word, EB_WORD_ALPHABET is returned.
 * Otherwise, -1 is returned.  It means that an error occurs.
 * `eb_error' is also set in this case.
 */
EB_Word_Code
eb_set_word(book, word, canonword, inputword)
    EB_Book *book;
    char *word;
    char *canonword;
    const char *inputword;
{
    EB_Word_Code wordcode;

    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return -1;
    }

    /*
     * Make a fixed word and a canonicalized word from `inputword'.
     */
    if (book->charcode == EB_CHAR_ISO8859_1) {
	wordcode = eb_fix_word_ltn(word, inputword);
	if (wordcode == EB_WORD_ERROR)
	    return EB_WORD_ERROR;
	strcpy(canonword, word);
    } else {
	wordcode = eb_fix_word_jis(word, inputword);
	if (wordcode == EB_WORD_ERROR)
	    return EB_WORD_ERROR;
	eb_canonicalize_word_jis(canonword, word);
    }

    return wordcode;
}


/*
 * Make a fixed word and a cannonicalized word for `ENDWORD SEARCH'.
 *
 * If `inputword' is a KANA word,  EB_WORD_KANA is retuend.
 * If `inputword' is a alphabetic word,  EB_WORD_ALPHABET is retuend.
 * Otherwise, -1 is returned.  It means that an error occurs.
 * `eb_error' is also set in this case.
 */
EB_Word_Code
eb_set_endword(book, word, canonword, inputword)
    EB_Book *book;
    char *word;
    char *canonword;
    const char *inputword;
{
    EB_Word_Code wordcode;

    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return -1;
    }

    /*
     * Make a fixed word and a canonicalized word from `inputword'.
     */
    if (book->charcode == EB_CHAR_ISO8859_1) {
	wordcode = eb_fix_word_ltn(word, inputword);
	if (wordcode == EB_WORD_ERROR)
	    return EB_WORD_ERROR;
	strcpy(canonword, word);
	eb_reverse_word_ltn(word);
	eb_reverse_word_ltn(canonword);
    } else {
	wordcode = eb_fix_word_jis(word, inputword);
	if (wordcode == EB_WORD_ERROR)
	    return EB_WORD_ERROR;
	eb_canonicalize_word_jis(canonword, word);
	eb_reverse_word_jis(word);
	eb_reverse_word_jis(canonword);
    }

    return wordcode;
}


/*
 * Convert `inputword' to ISO 8859 1 and put it into `word'.
 *
 * If `inputword' is a valid string to search, EB_WORD_ALPHA is returned.
 * Otherwise, -1 is returned.
 */
static int
eb_fix_word_ltn(word, inputword)
    char *word;
    const char *inputword;
{
    unsigned char *wp = (unsigned char *) word;
    const unsigned char *inp = (const unsigned char *) inputword;
    unsigned char c1;
    int keylen = 0;

    while (*inp != '\0') {
	/*
	 * Check for the length of the word.
	 * If exceeded, return with an error code.
	 */
	if (EB_MAXLEN_WORD < keylen + 1) {
	    eb_error = EB_ERR_TOO_LONG_WORD;
	    return EB_WORD_ERROR;
	}

	/*
	 * Tabs are translated to spaces.
	 */
	c1 = *inp++;
	if (c1 == ' ' || c1 == '\t')
	    continue;

	if ('a' <= c1 && c1 <= 'z')
	    *wp++ = c1 - 0x20;
	else
	    *wp++ = c1;

	keylen++;
    }
    *wp = '\0';

    if (keylen == 0) {
	eb_error = EB_ERR_EMPTY_WORD;
	return EB_WORD_ERROR;
    }

    return EB_WORD_ALPHA;
}


/*
 * Table used to convert JIS X 0208 to ASCII (subset of ISO 8859 1).
 */
static unsigned int jisx0208_table[] = {
    /* 0x20 -- 0x2f */
    0x2121, 0x212a, 0x2149, 0x2174, 0x2170, 0x2173, 0x2175, 0x2147,
    0x214a, 0x214b, 0x2176, 0x215c, 0x2124, 0x215d, 0x2125, 0x213f, 
    /* 0x30 -- 0x3f */
    0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, 0x2337, 
    0x2338, 0x2339, 0x2127, 0x2128, 0x2163, 0x2161, 0x2164, 0x2129, 
    /* 0x40 -- 0x4f */
    0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, 0x2347, 
    0x2348, 0x2349, 0x234a, 0x234b, 0x234c, 0x234d, 0x234e, 0x234f, 
    /* 0x50 -- 0x5f */
    0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, 0x2357, 
    0x2358, 0x2359, 0x235a, 0x214e, 0x2140, 0x214f, 0x2130, 0x2132, 
    /* 0x60 -- 0x6f */
    0x2146, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, 0x2367, 
    0x2368, 0x2369, 0x236a, 0x236b, 0x236c, 0x236d, 0x236e, 0x236f, 
    /* 0x70 -- 0x7e */
    0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, 0x2377, 
    0x2378, 0x2379, 0x237a, 0x2150, 0x2143, 0x2151, 0x2141, 
};

/*
 * Convert `inputword' to JIS X0208 and put it into `word'.
 *
 * If `inputword' is a valid string to search, EB_WORD_ALPHA or
 * EB_WORD_KANA is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_fix_word_jis(word, inputword)
    char *word;
    const char *inputword;
{
    unsigned char *wp = (unsigned char *) word;
    const unsigned char *inp = (const unsigned char *) inputword;
    unsigned char c1, c2;
    int kana = 0;
    int alphabet = 0;
    int kanji = 0;
    int keylen = 0;

    while (*inp != '\0') {
	/*
	 * Check for the length of the word.
	 * If exceeded, return with an error code.
	 */
	if (EB_MAXLEN_WORD < keylen + 2) {
	    eb_error = EB_ERR_TOO_LONG_WORD;
	    return EB_WORD_ERROR;
	}

	/*
	 * Tabs are translated to spaces.
	 */
	c1 = *inp++;
	if (c1 == '\t' || c1 == ' ')
	    continue;

	if (0x20 <= c1 && c1 <= 0x7e) {
	    unsigned int c = jisx0208_table[c1 - 0x20];
	    c1 = c >> 8;
	    c2 = c & 0xff;
	} else if (0xa1 <= c1 && c1 <= 0xfe) {
	    c2 = *inp++;
	    if (c2 < 0xa1 || 0xfe < c2) {
		eb_error = EB_ERR_BAD_WORD;
		return EB_WORD_ERROR;
	    }
	    if (c1 == 0xa1 && c2 == 0xa1)
		continue;
	    c1 &= 0x7f;
	    c2 &= 0x7f;
	} else {
	    eb_error = EB_ERR_BAD_WORD;
	    return EB_WORD_ERROR;
	}

	/*
	 * The following characters are recognized as alphabet.
	 *   2330 - 2339: `0' .. `9'
	 *   2341 - 235a: `A' .. `Z'
	 *   2361 - 237a: `a' .. `z' (convert to upper cases)
	 */
	if (c1 == 0x23 && 0x61 <= c2 && c2 <= 0x7a) {
	    *wp ++ = c1;
	    *wp ++ = c2 - 0x20;
	} else {
	    *wp++ = c1;
	    *wp++ = c2;
	}

	if (c1 == 0x23)
	    alphabet++;
	else if (c1 == 0x24 || c1 == 0x25)
	    kana++;
	else if (c1 != 0x21)
	    kanji++;

	keylen += 2;
    }
    *wp = '\0';

    if (keylen == 0) {
	eb_error = EB_ERR_EMPTY_WORD;
	return EB_WORD_ERROR;
    }
    if (kanji != 0)
	return EB_WORD_KANJI;
    else if (kana != 0)
	return EB_WORD_KANA;
    else
	return EB_WORD_ALPHA;

    /* not reached */
    return EB_WORD_ERROR;
}


/*
 * Table used to convert KATAKANA to HIRAGANA. 
 */
static const char
canon_table[] = {
    0x22, /* a(2321) -> A(2322) */	0x22, /* A(2322) -> A(2322) */
    0x24, /* i(2323) -> I(2324) */	0x24, /* I(2324) -> I(2324) */
    0x26, /* u(2325) -> U(2326) */	0x26, /* U(2326) -> U(2326) */
    0x28, /* e(2327) -> E(2328) */	0x28, /* E(2328) -> E(2328) */
    0x2a, /* o(2329) -> O(232a) */	0x2a, /* O(232a) -> O(232a) */
    0x2b, /* KA(232b) -> KA(232b) */	0x2b, /* GA(232c) -> KA(232b) */
    0x2d, /* KI(232d) -> KI(232d) */	0x2d, /* GI(232e) -> KI(232d) */
    0x2f, /* KU(232f) -> KU(232f) */	0x2f, /* GU(2330) -> KU(232f) */
    0x31, /* KE(2331) -> KE(2331) */	0x31, /* GE(2332) -> KE(2331) */
    0x33, /* KO(2333) -> KO(2333) */	0x33, /* GO(2334) -> KO(2333) */
    0x35, /* SA(2335) -> SA(2335) */	0x35, /* ZA(2336) -> SA(2335) */
    0x37, /* SI(2337) -> SI(2337) */	0x37, /* ZI(2338) -> SI(2337) */
    0x39, /* SU(2339) -> SU(2339) */	0x39, /* ZU(233a) -> SU(2339) */
    0x3b, /* SE(233b) -> SE(233b) */	0x3b, /* ZE(233c) -> SE(233b) */
    0x3d, /* SO(233d) -> SO(233d) */	0x3d, /* ZO(233e) -> SO(233d) */
    0x3f, /* TA(233f) -> TA(233f) */	0x3f, /* DA(2340) -> TA(233f) */
    0x41, /* TI(2341) -> TI(2341) */	0x41, /* DI(2342) -> TI(2341) */
    0x44, /* tu(2343) -> TU(2344) */	0x44, /* TU(2344) -> TU(2344) */
    0x44, /* DU(2345) -> TU(2344) */	0x46, /* TE(2346) -> TE(2346) */
    0x46, /* DE(2347) -> TE(2346) */	0x48, /* TO(2348) -> TO(2348) */
    0x48, /* DO(2349) -> TO(2348) */	0x4a, /* NA(234a) -> NA(234a) */
    0x4b, /* NI(234b) -> NI(234b) */	0x4c, /* NU(234c) -> NU(234c) */
    0x4d, /* NE(234d) -> NE(234d) */	0x4e, /* NO(234e) -> NO(234e) */
    0x4f, /* HA(234f) -> HA(234f) */	0x4f, /* BA(2350) -> HA(234f) */
    0x4f, /* PA(2351) -> HA(234f) */	0x52, /* HI(2352) -> HI(2352) */
    0x52, /* BI(2353) -> HI(2352) */	0x52, /* PI(2354) -> HU(2352) */
    0x55, /* HU(2355) -> HU(2355) */	0x55, /* BU(2356) -> HU(2355) */
    0x55, /* PU(2357) -> HU(2355) */	0x58, /* HE(2358) -> HE(2358) */
    0x58, /* BE(2359) -> HE(2358) */	0x58, /* PE(235a) -> HE(2358) */
    0x5b, /* HO(235b) -> HO(235b) */	0x5b, /* BO(235c) -> HO(235b) */
    0x5b, /* PO(235d) -> HO(235b) */	0x5e, /* MA(235e) -> MA(235e) */
    0x5f, /* MI(235f) -> MI(235f) */	0x60, /* MU(2360) -> MU(2360) */
    0x61, /* ME(2361) -> ME(2361) */	0x62, /* MO(2362) -> MO(2362) */
    0x64, /* ya(2363) -> YA(2364) */	0x64, /* YA(2364) -> YA(2364) */
    0x66, /* yu(2365) -> YU(2366) */	0x66, /* YU(2366) -> YU(2366) */
    0x68, /* yo(2367) -> YO(2368) */	0x68, /* YO(2368) -> YO(2368) */
    0x69, /* RA(2369) -> TA(2369) */	0x6a, /* RI(236a) -> RI(236a) */
    0x6b, /* RU(236b) -> RU(236b) */	0x6c, /* RE(236c) -> RE(236c) */
    0x6d, /* RO(236d) -> RO(236d) */	0x6e, /* wa(236e) -> WA(236e) */
    0x6f, /* WA(236f) -> WA(236f) */	0x70, /* WI(2370) -> WI(2370) */
    0x71, /* WE(2371) -> WE(2371) */	0x72, /* WO(2372) -> WO(2372) */
    0x73, /* N(2373) -> N(2373) */
};

/*
 * Table used to canonicalize a KANA word.
 */
static const char
vowel_table[] = {
    0x22, /* a(2321) -> A(2322) */	0x22, /* A(2322) -> A(2322) */
    0x24, /* i(2323) -> I(2324) */	0x24, /* I(2324) -> I(2324) */
    0x26, /* u(2325) -> U(2326) */	0x26, /* U(2326) -> U(2326) */
    0x28, /* e(2327) -> E(2328) */	0x28, /* E(2328) -> E(2328) */
    0x2a, /* o(2329) -> O(232a) */	0x2a, /* O(232a) -> O(232a) */
    0x22, /* KA(232b) -> A(2322) */	0x22, /* GA(232c) -> A(2322) */
    0x24, /* KI(232d) -> I(2324) */	0x24, /* GI(232e) -> I(2324) */
    0x26, /* KU(232f) -> U(2326) */	0x26, /* GU(2330) -> U(2326) */
    0x28, /* KE(2331) -> E(2328) */	0x28, /* GE(2332) -> E(2328) */
    0x2a, /* KO(2333) -> O(232a) */	0x2a, /* GO(2334) -> O(232a) */
    0x22, /* SA(2335) -> A(2322) */	0x22, /* ZA(2336) -> A(2322) */
    0x24, /* SI(2337) -> I(2324) */	0x24, /* ZI(2338) -> I(2324) */
    0x26, /* SU(2339) -> U(2326) */	0x26, /* ZU(233a) -> U(2326) */
    0x28, /* SE(233b) -> E(2328) */	0x28, /* ZE(233c) -> E(2328) */
    0x2a, /* SO(233d) -> O(232a) */	0x2a, /* ZO(233e) -> O(232a) */
    0x22, /* TA(233f) -> A(2322) */	0x22, /* DA(2340) -> A(2322) */
    0x24, /* TI(2341) -> I(2324) */	0x24, /* DI(2342) -> I(2324) */
    0x26, /* tu(2343) -> U(2326) */	0x26, /* TU(2344) -> U(2326) */
    0x26, /* DU(2345) -> U(2326) */	0x28, /* TE(2346) -> E(2328) */
    0x28, /* DE(2347) -> E(2328) */	0x2a, /* TO(2348) -> O(232a) */
    0x2a, /* DO(2349) -> O(232a) */	0x22, /* NA(234a) -> A(2322) */
    0x24, /* NI(234b) -> I(2324) */	0x26, /* NU(234c) -> U(2326) */
    0x28, /* NE(234d) -> E(2328) */	0x2a, /* NO(234e) -> O(232a) */
    0x22, /* HA(234f) -> A(2322) */	0x22, /* BA(2350) -> A(2322) */
    0x22, /* PA(2351) -> A(2322) */	0x24, /* HI(2352) -> I(2324) */
    0x24, /* BI(2353) -> I(2324) */	0x24, /* PI(2354) -> I(2324) */
    0x26, /* HU(2355) -> U(2326) */	0x26, /* BU(2356) -> U(2326) */
    0x26, /* PU(2357) -> U(2326) */	0x28, /* HE(2358) -> E(2328) */
    0x28, /* BE(2359) -> E(2328) */	0x28, /* PE(235a) -> E(2328) */
    0x2a, /* HO(235b) -> O(232a) */	0x2a, /* BO(235c) -> O(232a) */
    0x2a, /* PO(235d) -> O(232a) */	0x22, /* MA(235e) -> A(2322) */
    0x24, /* MI(235f) -> I(2324) */	0x26, /* MU(2360) -> U(2326) */
    0x28, /* ME(2361) -> E(2328) */	0x2a, /* MO(2362) -> O(232a) */
    0x22, /* ya(2363) -> A(2322) */	0x22, /* YA(2364) -> A(2322) */
    0x26, /* yu(2365) -> U(2326) */	0x26, /* YU(2366) -> U(2326) */
    0x2a, /* yo(2367) -> O(232a) */	0x2a, /* YO(2368) -> O(232a) */
    0x22, /* RA(2369) -> A(2322) */	0x24, /* RI(236a) -> I(2324) */
    0x26, /* RU(236b) -> U(2326) */	0x28, /* RE(236c) -> E(2328) */
    0x2a, /* RO(236d) -> O(232a) */	0x22, /* wa(236e) -> A(2322) */
    0x22, /* WA(236f) -> A(2322) */	0x24, /* WI(2370) -> I(2324) */
    0x28, /* WE(2371) -> E(2328) */	0x2a, /* WO(2372) -> O(232a) */
};


/*
 * Canonicalize `word' and put it into `canonword'.
 *
 * To search a KANA word in a subbook, a canocialized word is needed
 * in addition to the fixed word.
 * The canonicalized word is generated from `word' by the following
 * processes:
 *
 *   1. KATAKANA in the fixed word are converted to corresponding
 *      HIRAGANA.
 *   2. Voiced consonants, p-sounds, contracted sounds, and double
 *      consonants in the fixed word are converted to an unvoiced
 *      consonants.
 *   3. Vowel marks in the fixed word are converted to the previous
 *      HIRAGANA vowels.
 * 
 * Characters neither HIRAGANA nor KATAKANA are also put into `canonword'
 * without any conversion.
 */
static void
eb_canonicalize_word_jis(canonword, word)
    char *canonword;
    const char *word;
{
    unsigned char *cwp = (unsigned char *) canonword;
    const unsigned char *wp = (const unsigned char *) word;
    unsigned char c1, c2;
    unsigned char prevc1 = '\0', prevc2 = '\0';

    while (*wp != '\0' && *(wp + 1) != '\0') {
	c1 = *wp++;
	c2 = *wp++;

	if (c1 == 0x25 && 0x21 <= c2 && c2 <= 0x76) {
	    /*
	     * The character is a KATAKANA.
	     * Convert to HIRAGANA, and then convert to an unvoiced
	     * consonant.
	     *
	     * (0x2574 is `VU', 0x2575 is `ka', 0x2576 is `ke')
	     */
	    c1 = 0x24;
	    if (c2 == 0x74)
		c2 = 0x26;
	    if (c2 == 0x75)
		c2 = 0x2b;
	    else if (c2 == 0x76)
		c2 = 0x31;
	    else
		c2 = canon_table[c2 - 0x21];

	} else if (c1 == 0x23 && 0x61 <= c2 && c2 <= 0x7a) {
	    /*
	     * The character is an alphabet.  Convert to upper case.
	     */
	    c2 -= 0x20;

	} else if (c1 == 0x24 && 0x21 <= c2 && c2 <= 0x73) {
	    /*
	     * The character is a HIRAGANA.
	     * Convert to an unvoiced consonant.
	     */
	    c2 = canon_table[c2 - 0x21];

	} else if (c1 == 0x21 && c2 == 0x3c) {
	    /*
	     * The character is a long vowel mark.
	     * Convert to a HIRAGANA vowel of the previous KANA character.
	     * If previous character is not KANA, the conversion is not
	     * done.
	     */
	    if (prevc1 == 0x24 && 0x21 <= prevc2 && prevc2 <= 0x72) {
		c1 = 0x24;
		c2 = vowel_table[prevc2 - 0x21];
	    }
	}
	*cwp++ = c1;
	*cwp++ = c2;
	prevc1 = c1;
	prevc2 = c2;
    }
    *cwp = '\0';
}


/*
 * Reverse a word for ENDWORD SEARCH.
 * 
 * `word' is a word to reverse.  It must be an alphabetic word.
 * The reversed word is also put into `word'.
 */
static void
eb_reverse_word_ltn(word)
    char *word;
{
    char *p1, *p2;
    int len;
    char c;

    len = strlen(word);
    if (len == 0)
	return;
    for (p1 = word, p2 = word + len - 1; p1 < p2; p1++, p2--) {
	c = *p1;
	*p1 = *p2;
	*p2 = c;
    }
}


/*
 * Reverse a word for ENDWORD SEARCH.
 * 
 * `word' is a word to reverse.  It must be a KANA word.
 * The reversed word is also put into `word'.
 */
static void
eb_reverse_word_jis(word)
    char *word;
{
    char *p1, *p2;
    int len;
    char c;

    len = strlen(word);
    if (len % 2 == 1) {
	*(word + len - 1) = '\0';
	len--;
    }
    if (len < 2)
	return;
    
    for (p1 = word, p2 = word + len - 2; p1 < p2; p1 += 2, p2 -= 2) {
	c = *p1;
	*p1 = *p2;
	*p2 = c;
	c = *(p1 + 1);
	*(p1 + 1) = *(p2 + 1);
	*(p2 + 1) = c;
    }
}


