/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef EB_TYPES_H
#define EB_TYPES_H

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>

#ifdef EB_BUILD_LIBRARY
#include "defs.h"
#else
#include <eb/defs.h>
#endif


/*
 * Types for various codes.
 */
typedef int EB_Book_Code;
typedef int EB_Disc_Code;
typedef int EB_Case_Code;
typedef int EB_Suffix_Code;
typedef int EB_Character_Code;
typedef int EB_Font_Code;
typedef int EB_Word_Code;
typedef int EB_Language_Code;
typedef int EB_Message_Code;
typedef int EB_Subbook_Code;
typedef int EB_Index_Code;


/*
 * EB_Alternation_Cache -- An Aternation cache entry.
 */
typedef struct {
    int charno;
    char text[EB_MAXLEN_ALTERNATION_TEXT + 1];
} EB_Alternation_Cache;


/*
 * EB_Appendix_Subbook -- An appendix for a subbook.
 */
typedef struct {
    /*
     * Initialization flag.
     */
    int initialized;

    /*
     * Subbook ID.
     */
    EB_Subbook_Code code;

    /*
     * Directory name.
     */
    char directory[EB_MAXLEN_BASENAME + 1];

    /*
     * File descriptor for the appendix file.
     */
    int appfile;

    /*
     * Character code of the book.
     */
    EB_Character_Code charcode;

    /*
     * Start number of the character code in the narrow/wide font.
     */
    int narwstart;
    int widestart;

    /*
     * End number of the character code in the narrow/wide font.
     */
    int narwend;
    int wideend;

    /*
     * Start page number of the narrow/wide font.
     */
    int narwpage;
    int widepage;

    /*
     * Stop code.
     */
    int stop0;
    int stop1;
} EB_Appendix_Subbook;


/*
 * EB_Appendix -- Additional resources for a book.
 */
typedef struct {
    /*
     * Path of the book.
     */
    char *path;

    /*
     * The length of the path.
     */
    size_t pathlen;

    /*
     * Disc type.  EB(EB/EBG/EBXA) or EPWING.
     */
    EB_Disc_Code disccode;

    /*
     * Cases of the filenames; upper or lower.
     */
    EB_Case_Code casecode;

    /*
     * Suffix to be added to filenames. (None, ".", or ".;1")
     */
    EB_Suffix_Code suffixcode;

    /*
     * The number of subbooks the book has.
     */
    int subcount;

    /*
     * Subbook list.
     */
    EB_Appendix_Subbook *subbooks;

    /*
     * Current subbook.
     */
    EB_Appendix_Subbook *subcurrent;

    /*
     * Cache table for alternation text.
     */
    EB_Alternation_Cache narwcache[EB_MAX_ALTERNATION_CACHE];
    EB_Alternation_Cache widecache[EB_MAX_ALTERNATION_CACHE];
} EB_Appendix;


/*
 * EB_Font -- A font in a subbook.
 */
typedef struct {
    /* 
     * Width and height.
     */
    int width;
    int height;

    /*
     * Character numbers of the start and end of the font.
     */
    int start;
    int end;

    /*
     * Page number of the start page of the font data.
     * (EB/EBG/EBXA only.  In EPWING, it is alyways 1).
     */
    int page;

    /*
     * File descriptor for the font file. (EPWING only)
     */
    int fontfile;

    /*
     * Filename of the font. (EPWING only)
     */
    char filename[EB_MAXLEN_BASENAME + 1];
} EB_Font;


/*
 * EB_Language -- A language in a book. (EB/EBG/EBXA only)
 */
typedef struct {
    /*
     * Language ID.
     */
    EB_Language_Code code;

    /*
     * Offset of the messages in the language file.
     */
    off_t offset;

    /*
     * The number of messages the language file has.
     */
    int msgcount;

    /*
     * Language name.
     */
    char name[EB_MAXLEN_LANGNAME + 1];
} EB_Language;


/*
 * EB_Subbook -- A subbook in the book.
 */
typedef struct {
    /*
     * Whether information about the subbook have been loaded in the
     * structure.
     */
    int initialized;

    /*
     * Subbook ID.
     */
    EB_Subbook_Code code;

    /*
     * File descriptor for the subbook file.
     */
    int startfile;

    /*
     * Title of the subbook.
     */
    char title[EB_MAXLEN_TITLE + 1];

    /*
     * Directory name.
     */
    char directory[EB_MAXLEN_BASENAME + 1];

    /*
     * The page numbers of search methods.
     */
    int indexes[EB_MAX_INDEXES];

    /*
     * The number of fonts the subbook has.
     */
    int fontcount;

    /*
     * Font list.
     */
    EB_Font fonts[EB_MAX_FONTS * 2];

    /*
     * Current narrow and wide fonts.
     */
    EB_Font *narwcurrent;
    EB_Font *widecurrent;
} EB_Subbook;


/*
 * EB_Book -- A book.
 */
typedef struct {
    /*
     * Book ID.
     */
    EB_Book_Code code;

    /*
     * Disc type.  EB(EB/EBG/EBXA) or EPWING.
     */
    EB_Disc_Code disccode;

    /*
     * Character code of the book.
     */
    EB_Character_Code charcode;

    /*
     * Path of the book.
     */
    char *path;

    /*
     * The length of the path.
     */
    size_t pathlen;

    /*
     * Cases of the filenames; upper or lower.
     */
    EB_Case_Code casecode;

    /*
     * Suffix to be added to filenames. (None, ".", or ".;1")
     */
    EB_Suffix_Code suffixcode;

    /*
     * The number of subbooks the book has.
     */
    int subcount;

    /*
     * Subbook list.
     */
    EB_Subbook *subbooks;

    /*
     * Current subbook.
     */
    EB_Subbook *subcurrent;

    /*
     * The number of languages the book supports.
     */
    int langcount;

    /*
     * Language list.
     */
    EB_Language *languages;

    /*
     * Current language.
     */
    EB_Language *langcurrent;

    /*
     * Messages in the current language.
     */
    char *messages;
} EB_Book;


/*
 * EB_Position -- A pair of page and offset.
 */
typedef struct {
    int page;
    int offset;
} EB_Position;


/*
 * EB_Hit -- In a word search, heading and text locations of a matched
 * entry are stored.
 */
typedef struct {
    EB_Position heading;
    EB_Position text;
} EB_Hit;


#ifdef __cplusplus
}
#endif

#endif /* not EB_TYPES_H */

